#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""Client that connects to a remote running Brayns instance which provides the supported API."""

import rockets

from .base import BaseClient
from .utils import build_schema_requests_from_registry, convert_snapshot_response_to_PIL


class Client(BaseClient):
    """Client that connects to a remote running Brayns instance which provides the supported API."""

    def __init__(self, url, loop=None):
        """
        Create a new client instance by connecting to the given URL.

        :param str url: a string 'hostname:port' to connect to a running Brayns instance
        :param asyncio.AbstractEventLoop loop: Event loop where this client should run in
        """
        super().__init__(url)

        self.rockets_client = rockets.Client(url, subprotocols=['rockets'], loop=loop)

        registry, requests = build_schema_requests_from_registry(self.http_url)
        schemas = self.rockets_client.batch(requests)
        super()._build_api(registry, requests, schemas)

    # pylint: disable=W0613,W0622,E1101
    def image(self, size, format='jpg', animation_parameters=None, camera=None, quality=None,
              renderer=None, samples_per_pixel=None):
        """
        Request a snapshot from Brayns and return a PIL image.

        :param tuple size: (width,height) of the resulting image
        :param str format: image type as recognized by FreeImage
        :param object animation_parameters: animation params to use instead of current params
        :param object camera: camera to use instead of current camera
        :param int quality: compression quality between 1 (worst) and 100 (best)
        :param object renderer: renderer to use instead of current renderer
        :param int samples_per_pixel: samples per pixel to increase render quality
        :return: the PIL image of the current rendering, None on error obtaining the image
        :rtype: :py:class:`~PIL.Image.Image`
        """
        args = locals()
        del args['self']
        result = self.snapshot(**{k: v for k, v in args.items() if v})
        return convert_snapshot_response_to_PIL(result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""Client that connects to a remote running Brayns instance which provides the supported API."""

import asyncio
import inspect
import rockets

from .api_generator import build_api
from .utils import base64decode, set_http_protocol, underscorize
from .utils import HTTP_METHOD_GET, HTTP_STATUS_OK
from .version import MINIMAL_VERSION
from . import utils


class BaseClient:
    """Client that connects to a remote running Brayns instance which provides the supported API."""

    def __init__(self, url):
        """
        Create a new client instance by connecting to the given URL.

        :param str url: a string 'hostname:port' to connect to a running Brayns instance
        """
        self.http_url = set_http_protocol(url) + '/'
        """The HTTP URL"""

        self._check_version()
        self.rockets_client = None

        if utils.in_notebook():
            self._add_widgets()  # pragma: no cover

    # pylint: disable=E1101
    def __str__(self):
        """Return a pretty-print on the currently connected Brayns instance."""
        version = 'unknown'
        if self.version:
            version = '.'.join(str(x) for x in [self.version.major, self.version.minor,
                                                self.version.patch])
        return "Brayns version {0} running on {1}".format(version, self.http_url)

    def set_colormap(self, palette, intensity=1, opacity=1, data_range=(0, 255)):
        """
        Set a colormap to Brayns.

        :param list palette: color palette to use
        :param int intensity: value to amplify the color values
        :param int opacity: opacity for colormap values
        :param typle data_range: data range on which values the colormap should be applied
        """
        palette_size = len(palette)
        contributions = []
        diffuses = []
        # pylint: disable=E1101
        tf = self.transfer_function
        for i in range(0, palette_size):
            color = palette[i]
            diffuses.append([intensity * color[0], intensity * color[1], intensity * color[2],
                             opacity])
            contributions.append(0)
        tf.diffuse = diffuses
        tf.contribution = contributions
        tf.range = data_range
        tf.commit()

    def open_ui(self):
        """Open the Brayns UI in a new page of the default system browser."""
        import webbrowser
        url = 'https://bbp-brayns.epfl.ch?host=' + self.http_url
        webbrowser.open(url)

    def _check_version(self):
        """Check if the Brayns' version is sufficient enough."""
        status = utils.http_request(HTTP_METHOD_GET, self.http_url, 'version')
        if status.code != HTTP_STATUS_OK:
            raise Exception('Cannot obtain version from Brayns')

        version = '.'.join(str(x) for x in [status.contents['major'], status.contents['minor'],
                                            status.contents['patch']])

        import semver
        if semver.match(version, '<{0}'.format(MINIMAL_VERSION)):
            raise Exception('Brayns does not satisfy minimal required version; '
                            'needed {0}, got {1}'.format(MINIMAL_VERSION, version))

    def _build_api(self, registry, requests, schemas):
        """Use the schemas from the remote running Brayns to build the API."""
        schemas_dict = dict()
        for request in requests:
            # pylint: disable=protected-access,cell-var-from-loop
            response = list(filter(lambda x, request_id=request.request_id(): x._id == request_id,
                                   schemas))
            # pylint: enable=protected-access,cell-var-from-loop
            schemas_dict[request.params['endpoint']] = response[0].result
        build_api(self, registry, schemas_dict)

        self._setup_notifications()

    def _setup_notifications(self):
        def _notification_filter(data):
            return data.method.startswith('set-')

        def _on_notification(data):
            # remove the 'set-' part of the data method to find the property to update its content
            prop_name = underscorize(data.method[4:])
            prop = getattr(self, '_' + prop_name, None)
            if prop:
                prop.update(data.params)

        self.rockets_client.notifications.filter(_notification_filter).subscribe(_on_notification)

    def _add_widgets(self):  # pragma: no cover
        """Add functions to the Brayns object to provide widgets for appropriate properties."""
        self._add_show_function()
        self._add_animation_slider()

    def _add_show_function(self):  # pragma: no cover
        """Add show() function for live streaming."""
        def function_builder():
            """Wrapper for returning the show() function."""
            def show():
                """Show the live rendering of Brayns."""
                from IPython.display import display
                import ipywidgets as widgets
                from rx import Observer

                rockets_client = self.rockets_client
                brayns = self

                class ImageStreamObserver(Observer):
                    """Update a Image widget with the JPEG stream from Brayns."""

                    def __init__(self):
                        self.image = widgets.Image(format='jpg')

                        def _show_image(value):
                            self.image.value = base64decode(value['data'])
                            viewport = brayns.application_parameters.viewport
                            self.image.width = viewport[0]
                            self.image.height = viewport[1]
                            display(self.image)

                        if isinstance(rockets_client, rockets.AsyncClient):
                            task = rockets_client.async_request('image-jpeg')
                            task.add_done_callback(lambda x: _show_image(x.result()))
                        else:
                            _show_image(rockets_client.request('image-jpeg'))

                    def on_next(self, value):
                        viewport = brayns.application_parameters.viewport
                        self.image.width = viewport[0]
                        self.image.height = viewport[1]
                        self.image.value = value

                    def on_completed(self):
                        self.image.close()

                    def on_error(self, error):
                        print("Error occurred: {0}".format(error))
                        self.image.close()

                rockets_client.ws_observable \
                    .filter(lambda value: isinstance(value, (bytes, bytearray, memoryview))) \
                    .subscribe(ImageStreamObserver())

            return show

        setattr(self, 'show', function_builder())

    def _add_animation_slider(self):  # pragma: no cover
        """Add animation_slider() function for animation_parameters control."""
        def function_builder():
            """Wrapper for returning the animation_slider() function."""
            def animation_slider():
                """Show slider to control animation."""
                from IPython.display import display
                import ipywidgets as widgets
                from rx import Observer

                brayns = self

                class AnimationObserver(Observer):
                    """Update a animation slider from 'set-animation-parameters' notifications."""

                    def __init__(self):
                        # pylint: disable=E1101
                        anim_params = brayns.animation_parameters
                        self.button = widgets.ToggleButton(value=anim_params.delta != 0,
                                                           icon='play' if anim_params.delta == 0
                                                           else 'pause')
                        self.button.layout = widgets.Layout(width='40px')

                        def _on_button_change(change):
                            self.button.icon = 'pause' if change['new'] else 'play'
                            result = brayns.set_animation_parameters(delta=1 if change['new']
                                                                     else 0)
                            if inspect.iscoroutine(result):
                                asyncio.ensure_future(result)

                        self.button.observe(_on_button_change, names='value')

                        self.slider = widgets.IntSlider(min=anim_params.start,
                                                        max=anim_params.end,
                                                        value=anim_params.current)

                        def _on_value_change(change):
                            result = brayns.set_animation_parameters(current=change['new'])
                            if inspect.iscoroutine(result):
                                asyncio.ensure_future(result)

                        self.value_change = _on_value_change

                        self.slider.observe(self.value_change, names='value')

                        self.box = widgets.HBox([self.button, self.slider])
                        display(self.box)

                    def on_next(self, value):
                        self.slider.unobserve(self.value_change, names='value')
                        params = value.params
                        self.slider.min = params['start']
                        self.slider.max = params['end']
                        self.slider.value = params['current']
                        self.slider.observe(self.value_change, names='value')

                    def on_completed(self):
                        self.box.close()

                    def on_error(self, error):
                        print("Error occurred: {0}".format(error))
                        self.box.close()

                def _animation_filter(value):
                    return value.method == 'set-animation-parameters'
                self.rockets_client.notifications.filter(_animation_filter) \
                    .subscribe(AnimationObserver())

            return animation_slider

        setattr(self, 'animation_slider', function_builder())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""Client that connects to a remote running Brayns instance which provides the supported API."""

import asyncio
import rockets

from .base import BaseClient
from .utils import build_schema_requests_from_registry, convert_snapshot_response_to_PIL
from . import utils


class aobject:
    """Inheriting from this class allows to define an async __init__."""

    async def __new__(cls, *a, **kw):
        """Allow to create objects by calling `await MyClass(params)`"""
        instance = super().__new__(cls)
        await instance.__init__(*a, **kw)
        return instance


class AsyncClient(BaseClient, aobject):
    """Client that connects to a remote running Brayns instance which provides the supported API."""

    async def __init__(self, url, loop=None):
        """
        Create a new client instance by connecting to the given URL.

        :param str url: a string 'hostname:port' to connect to a running Brayns instance
        :param asyncio.AbstractEventLoop loop: Event loop where this client should run in
        """
        super().__init__(url)

        self.rockets_client = rockets.AsyncClient(url, subprotocols=['rockets'], loop=loop)

        registry, requests = build_schema_requests_from_registry(self.http_url)
        schemas = await self.rockets_client.batch(requests)
        super()._build_api(registry, requests, schemas)

    # pylint: disable=W0613,W0622,E1101
    def image(self, size, format='jpg', animation_parameters=None, camera=None, quality=None,
              renderer=None, samples_per_pixel=None):
        """
        Request a snapshot from Brayns and return a PIL image.

        :param tuple size: (width,height) of the resulting image
        :param str format: image type as recognized by FreeImage
        :param object animation_parameters: animation params to use instead of current params
        :param object camera: camera to use instead of current camera
        :param int quality: compression quality between 1 (worst) and 100 (best)
        :param object renderer: renderer to use instead of current renderer
        :param int samples_per_pixel: samples per pixel to increase render quality
        :return: the asyncio.Future of the snapshot generation, or None if inside a notebook
        :rtype: :py:class:`~asyncio.Future`
        """
        args = locals()
        del args['self']
        result = self.snapshot(**{k: v for k, v in args.items() if v})

        future = asyncio.get_event_loop().create_future()

        def _on_done(task):
            try:
                if task.exception():  # pragma: no cover
                    print("image() failed:", task.exception())
                else:
                    image = convert_snapshot_response_to_PIL(task.result())
                    if utils.in_notebook():  # pragma: no cover
                        if image:
                            from IPython.display import display
                            display(image)
                    else:
                        future.set_result(image)
            except rockets.RequestError as e:  # pragma: no cover
                print("Error", e.code, e.message)
            except ConnectionRefusedError as e:  # pragma: no cover
                print(e)

        result.add_done_callback(_on_done)
        if utils.in_notebook():  # pragma: no cover
            return None
        return future



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""
Provides functionality to dynamically add properties, types and functions to a given object.

The information for the API generation is based on a registry and JSON-schema of exposed properties
and RPCs.
"""

import asyncio
import os
import inflection
import python_jsonschema_objects as pjs

import rockets

from .utils import HTTP_METHOD_PUT, SCHEMA_ENDPOINT, add_method, add_progress_cancel_widget, \
                   underscorize


def build_api(target_object, registry, schemas):
    """
    Add the API (types, properties, methods) to the given object from registry and schemas.

    :param object target_object: The target object where to add all generated properties, RPCs and
                                 types to
    :param dict registry: The registry of all endpoints which are properties and methods
    :param dict schemas: The schemas matching the endpoints from the registry
    """
    for registry_entry in registry.keys():
        if registry_entry in schemas:
            schema = schemas[registry_entry]
        elif registry_entry.endswith(SCHEMA_ENDPOINT):
            method_name = registry_entry[:-len(SCHEMA_ENDPOINT)]
            schema = schemas[method_name]

        if 'type' not in schema:
            continue

        if 'method' in schema['type']:
            _add_method(target_object, schema)
            continue
        if registry_entry.endswith(SCHEMA_ENDPOINT):
            continue

        writeable = HTTP_METHOD_PUT in registry[registry_entry]
        _try_add_property(target_object, registry_entry, schema, writeable)


def _try_add_property(target_object, registry_entry, schema, writeable):
    """
    Try to add registry_entry as a property.

    This will add a property with the name registry_entry which initializes itself with current
    value from Brayns on first access. Furthermore, it adds potential enum values as string
    constants and adds a commit() function if the property is writeable.

    :param object target_object: The target object where to add the property to
    :param str registry_entry: registry endpoint, e.g. foo[/schema]
    :param dict schema: The schema of the object behind the endpoint.
    :param bool writeable: if the property is writeable or not
    """
    classes = pjs.ObjectBuilder(schema).build_classes()
    class_names = dir(classes)

    # find the generated class name that matches our schema root title
    for c in class_names:
        if c.lower() == schema['title'].lower():
            class_name = c
            break

    class_type = getattr(classes, class_name)

    is_object = schema['type'] == 'object'
    if is_object:
        value = class_type()
        _add_enums(value, target_object)
        if writeable:
            _add_commit(target_object.rockets_client, class_type, registry_entry)
    else:  # array
        value = class_type(())

    # add member and property to target_object
    member = '_' + underscorize(os.path.basename(registry_entry))
    setattr(target_object, member, value)
    _add_property(target_object, member, registry_entry, schema['type'] == 'array')


def _make_arg_list(param):
    required = param['required'] if 'required' in param else list()
    optional = list()
    for s in param['properties'].keys():
        if s not in required:
            optional.append(s + '=None')
    arg_list = ', '.join(required)
    if optional and required:
        arg_list += ', '
    arg_list += ', '.join(filter(None, optional))
    return arg_list


def _add_method_with_object_arg(target_object, param, method, description, is_async, is_request):
    """
    Add a method to cls where each property of the param object is a key-value argument.

    :param object target_object: The target object where to add the method to
    :param dict param: the parameter object from the RPC
    :param str method: the name of RPC
    :param str description: the description of RPC
    :param bool is_async: if the RPC is processed asynchronously or not
    :param bool is_request: if the RPC is a request with a response or a notification only
    """
    arg_list = _make_arg_list(param)
    func_name = str(underscorize(os.path.basename(method)))
    cls = target_object.__class__
    params = 'params={k:v for k,v in args.items() if v is not None}'
    if is_request:
        if isinstance(target_object.rockets_client, rockets.AsyncClient):
            if is_async:
                code = '''
                    def function(self, {0}):
                        args = locals()
                        del args['self']
                        return self.rockets_client.async_request("{1}", {2})
                    '''.format(arg_list, method, params)
            else:
                code = '''
                    async def function(self, {0}):
                        args = locals()
                        del args['self']
                        return await self.rockets_client.request("{1}", {2})
                    '''.format(arg_list, method, params)
        else:
            code = '''
                def function(self, {0}, response_timeout=None):
                    args = locals()
                    del args['self']
                    return self.rockets_client.request("{1}", {2})
                '''.format(arg_list, method, params)
    else:
        if isinstance(target_object.rockets_client, rockets.AsyncClient):
            code = '''
                async def function(self, {0}):
                    args = locals()
                    del args['self']
                    return await self.rockets_client.notify("{1}", {2})
                '''.format(arg_list, method, params)
        else:
            code = '''
                def function(self, {0}):
                    args = locals()
                    del args['self']
                    return self.rockets_client.notify("{1}", {2})
                '''.format(arg_list, method, params)

    d = {}
    exec(code.strip(), d)  # pylint: disable=W0122
    function = d['function']
    add_method(cls, func_name, description)(add_progress_cancel_widget(function))


def _add_method_with_array_arg(target_object, param, method, description, is_async, is_request):
    """
    Add a method to cls where the argument is an array.

    :param object target_object: The target object where to add the method to
    :param dict param: the parameter object from the RPC
    :param str method: the name of RPC
    :param str description: the description of RPC
    :param bool is_async: if the RPC is processed asynchronously or not
    :param bool is_request: if the RPC is a request with a response or a notification only
    """
    if 'description' in param:
        description += '\n:param {0}: {1}'.format(param['name'], param['description'])
    func_name = str(underscorize(os.path.basename(method)))
    cls = target_object.__class__
    if is_request:
        if isinstance(target_object.rockets_client, rockets.AsyncClient):
            if is_async:
                @add_method(cls, func_name, description)
                @add_progress_cancel_widget
                def function(self, array):  # pylint: disable=C0111,W0612
                    return self.rockets_client.async_request(method, array)
            else:
                @add_method(cls, func_name, description)
                @add_progress_cancel_widget
                async def function(self, array):  # pylint: disable=C0111,W0612
                    return await self.rockets_client.request(method, array)
        else:
            @add_method(cls, func_name, description)
            def function(self, array, response_timeout=None):  # pylint: disable=C0111
                return self.rockets_client.request(method, array, response_timeout)
    else:
        if isinstance(target_object.rockets_client, rockets.AsyncClient):
            @add_method(cls, func_name, description)
            async def function(self, array):  # pylint: disable=C0111
                await self.rockets_client.notify(method, array)
        else:
            @add_method(cls, func_name, description)
            def function(self, array):  # pylint: disable=C0111
                self.rockets_client.notify(method, array)


def _add_method_with_oneof_arg(target_object, method, description, is_async, is_request):
    """
    Add a method to cls where the argument is from a oneOf array.

    :param object target_object: The target object where to add the oneOf types to
    :param str method: name of RPC
    :param str description: description of RPC
    :param bool is_async: if the RPC is processed asynchronously or not
    :param bool is_request: if the RPC is a request with a response or a notification only
    """
    func_name = str(underscorize(os.path.basename(method)))
    cls = target_object.__class__
    if is_request:
        if isinstance(target_object.rockets_client, rockets.AsyncClient):
            if is_async:
                @add_method(cls, func_name, description)
                @add_progress_cancel_widget
                def function(self, params):  # pylint: disable=C0111,W0612
                    return self.rockets_client.async_request(method, params.for_json())
            else:
                @add_method(cls, func_name, description)
                @add_progress_cancel_widget
                async def function(self, params):  # pylint: disable=C0111,W0612
                    return await self.rockets_client.request(method, params.for_json())
        else:
            @add_method(cls, func_name, description)
            def function(self, params, response_timeout=None):  # pylint: disable=C0111
                return self.rockets_client.request(method, params.for_json(), response_timeout)
    else:
        if isinstance(target_object.rockets_client, rockets.AsyncClient):
            @add_method(cls, func_name, description)
            async def function(self, params):  # pylint: disable=C0111
                await self.rockets_client.notify(method, params.for_json())
        else:
            @add_method(cls, func_name, description)
            def function(self, params):  # pylint: disable=C0111
                self.rockets_client.notify(method, params.for_json())


def _add_method_with_no_args(target_object, method, description, is_async, is_request):
    """
    Add a method to cls which has no argument.

    :param object target_object: The target object where to add the method to
    :param str method: name of RPC
    :param str description: description of RPC
    :param bool is_async: if the RPC is processed asynchronously or not
    :param bool is_request: if the RPC is a request with a response or a notification only
    """
    func_name = str(underscorize(os.path.basename(method)))
    cls = target_object.__class__
    if is_request:
        if isinstance(target_object.rockets_client, rockets.AsyncClient):
            if is_async:
                @add_method(cls, func_name, description)
                @add_progress_cancel_widget
                def _function(self):
                    return self.rockets_client.async_request(method)
            else:
                @add_method(cls, func_name, description)
                async def _function(self):
                    return await self.rockets_client.request(method)
        else:
            @add_method(cls, func_name, description)
            def _function(self, response_timeout=None):
                return self.rockets_client.request(method, response_timeout=response_timeout)
    else:
        if isinstance(target_object.rockets_client, rockets.AsyncClient):
            @add_method(cls, func_name, description)
            async def _function(self):  # pylint: disable=C0111
                await self.rockets_client.notify(method)
        else:
            @add_method(cls, func_name, description)
            def _function(self):  # pylint: disable=C0111
                self.rockets_client.notify(method)


def _add_enums_from_oneof_types(target_object, param, method):
    """
    Create a type for each oneOf type in param and add it to target_object.

    :param object target_object: The target object where to add the oneOf types to
    :param list param: the oneOf array
    :param str method: name of RPC
    """
    for o in param:
        classes = pjs.ObjectBuilder(o).build_classes()
        class_names = dir(classes)

        # find the generated class name that matches our schema root title
        class_name = None
        for c in class_names:
            if c.lower() == inflection.camelize(o['title'].replace(' ', '_')).lower():
                class_name = c
                break

        # create class name <Type><Method w/o set->, e.g. Perspective+CameraParams
        pretty_class_name = underscorize(method[4:])
        pretty_class_name = inflection.camelize(pretty_class_name)
        pretty_class_name = class_name + pretty_class_name

        # add type to target_object
        class_type = getattr(classes, class_name)
        setattr(target_object, pretty_class_name, class_type)

        # create and add potential enums to type
        _add_enums(class_type(), class_type)


def _add_enums(root_object, target_object):
    """
    Look for enums in the given object to create string constants <ENUM_CLASSNAME>_<ENUM_VALUE>.

    :param dict root_object: the object which may contain enums
    :param object target_object: The target object where to add the string constants to
    """
    for i in root_object.keys():
        enum = None
        propinfo = root_object.propinfo(i)
        if 'enum' in propinfo:
            enum = propinfo
        if propinfo['type'] == 'array':
            if 'enum' in propinfo['items']:
                enum = propinfo['items']
        if not enum:
            continue

        enum_class = str(i).upper()
        if 'title' in enum:
            enum_class = str(inflection.underscore(enum['title'])).upper()
        enum_class += "_"
        for val in enum['enum']:
            enum_value = enum_class + inflection.parameterize(val, '_').upper()
            setattr(target_object, enum_value, val)


def _add_method(target_object, schema):
    """
    Add a new method to target_object from the given schema that describes an RPC.

    :param object target_object: The target object where to add the method to
    :param dict schema: schema containing name, description, params of the RPC
    :raises Exception: if the param type of the RPC does not match oneOf, object or array
    """
    method = schema['title']

    if 'params' in schema and len(schema['params']) > 1:
        print("Multiple arguments for RPC '{0}' not supported".format(method))
        return

    description = schema['description']
    is_async = schema['async']
    is_request = 'returns' in schema and schema['returns']

    if 'params' in schema and len(schema['params']) == 1:
        params = schema['params'][0]
        if 'oneOf' in params:
            _add_enums_from_oneof_types(target_object, params['oneOf'], method)
            _add_method_with_oneof_arg(target_object, method, description, is_async, is_request)
        # in the absence of multiple argument support, create a function with multiple arguments
        # from object properties
        elif params['type'] == 'object':
            _add_method_with_object_arg(target_object, params, method, description,
                                        is_async, is_request)
        elif params['type'] == 'array':
            _add_method_with_array_arg(target_object, params, method, description,
                                       is_async, is_request)
        else:
            raise Exception('Invalid argument type for method "{0}": '.format(method) +
                            'must be "object", "array" or "oneOf"')
    else:
        _add_method_with_no_args(target_object, method, description, is_async, is_request)


def _add_commit(rpc_client, property_type, object_name):
    """Add commit() for given property type."""
    def commit_builder(url):
        """Wrapper for returning the property.commit() function."""
        def commit(prop):
            """Update the property in the remote side."""
            return rpc_client.request('set-' + os.path.basename(url), prop.as_dict())

        return commit

    setattr(property_type, 'commit', commit_builder(object_name))


def _add_property(target_object, member, property_name, is_array):
    """Add property to target_object which initializes itself on first access."""
    def getter_builder(member, property_name):
        """Wrapper for returning the property state."""
        def function(self):
            """Returns the current state for the property."""
            value = getattr(self, member)

            # Initialize on first access; updates are received via websocket
            if is_array:
                has_value = value.data
            else:
                has_value = value.as_dict()

            rockets_client = target_object.rockets_client
            if not has_value or not rockets_client.connected():
                def _init_value(value, new_value, is_array):
                    if 'code' not in new_value:
                        if is_array:
                            value.__init__(new_value)
                        else:
                            value.__init__(**new_value)

                GET_PROPERTY = 'get-'+property_name
                if isinstance(rockets_client, rockets.AsyncClient):
                    task = new_value = rockets_client.async_request(GET_PROPERTY)
                    task.add_done_callback(lambda task: _init_value(value, task.result(), is_array))
                    loop = asyncio.get_event_loop()
                    if not loop.is_running():
                        loop.run_until_complete(task)
                else:
                    new_value = rockets_client.request(GET_PROPERTY)
                    _init_value(value, new_value, is_array)

            if is_array:
                return value.data
            return value

        return function

    endpoint_name = os.path.basename(property_name)
    snake_case_name = underscorize(endpoint_name)
    setattr(type(target_object), snake_case_name,
            property(fget=getter_builder(member, property_name),
                     doc='Access to the {0} property'.format(endpoint_name)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2018, Blue Brain Project
#                          Raphael Dumusc <raphael.dumusc@epfl.ch>
#                          Daniel Nachbaur <daniel.nachbaur@epfl.ch>
#                          Cyrille Favreau <cyrille.favreau@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# All rights reserved. Do not distribute without further notice.

"""Utils module for http request and notebook stuff."""

import base64
import io
import sys
from collections import OrderedDict
from functools import wraps
from PIL import Image
import requests
import rockets


HTTP_METHOD_PUT = 'PUT'
HTTP_METHOD_GET = 'GET'
HTTP_METHOD_DELETE = 'DELETE'
HTTP_METHOD_POST = 'POST'
HTTP_STATUS_OK = 200

HTTP_PREFIX = 'http://'
HTTPS_PREFIX = 'https://'

WS_PATH = '/ws'

SCHEMA_ENDPOINT = '/schema'


class Status:
    """Holds the execution status of an HTTP request."""

    def __init__(self, code, contents):
        """Initialize status object with given code and contents."""
        self.code = code
        self.contents = contents


# pylint: disable=R0912
def http_request(method, url, command, body=None, query_params=None):  # pragma: no cover
    """
    Perform http requests to the given URL and return the applications' response.

    :param str method: the type of HTTP request, PUT or GET are supported
    :param str url: the URL of the applications' http server
    :param str body: optional body for PUT requests
    :param str command: the type of HTTP command to be executed on the target app
    :param str query_params: the query params to append to the request url
    :return: JSON-encoded response of the request
    :rtype: Status
    :raises Exception: on connection error
    """
    full_url = url
    request = None
    full_url += command
    try:
        if method == HTTP_METHOD_POST:
            if body == '':
                request = requests.post(full_url, params=query_params)
            else:
                request = requests.post(full_url, data=body, params=query_params)
        elif method == HTTP_METHOD_PUT:
            if body == '':
                request = requests.put(full_url, params=query_params)
            else:
                request = requests.put(full_url, data=body, params=query_params)
        elif method == HTTP_METHOD_GET:
            request = requests.get(full_url, params=query_params)
            if request.status_code == 502:
                raise requests.exceptions.ConnectionError('Bad Gateway 502')
        elif method == HTTP_METHOD_DELETE:
            if body == '':
                request = requests.delete(full_url, params=query_params)
            else:
                request = requests.delete(full_url, data=body, params=query_params)
        js = ''
        if request.content:
            if request.status_code == 200:
                js = request.json(object_pairs_hook=OrderedDict)
            else:
                js = request.text
        response = Status(request.status_code, js)
        request.close()
    except requests.exceptions.ConnectionError:
        raise Exception('ERROR: Failed to connect to Brayns, did you start it with the '
                        '--http-server command line option?')
    return response


def in_notebook():
    """
    Check we are running in a notebook or not.

    :return: ``True`` if the module is running in IPython kernel, ``False`` if in IPython shell or
    other Python shell.
    :rtype: bool
    """
    return 'ipykernel' in sys.modules


def set_http_protocol(url):
    """
    Set the http protocol to the url if it is not present.

    :param str url: Url to be checked
    :return: Url with protocol
    :rtype: str
    """
    if url.find(HTTP_PREFIX) == -1 and url.find(HTTPS_PREFIX) == -1:
        return HTTP_PREFIX + url
    return url


def underscorize(word):
    """
    Opposite of inflection.dasherize; replace dashes with underscore in the word.

    :param str word: the word to underscorize
    :return: the underscorized word
    :rtype: str
    """
    return word.replace('-', '_')


def add_method(cls, name, description):
    """
    Decorator that adds the decorated function with the given name and docstring to cls.

    :param object cls: the python class to the decorated function to
    :param str name: the name of the function
    :param str description: the docstring of the function
    :return: the decorator
    :rtype: decorator
    """
    def _decorator(func):
        func.__doc__ = description

        @wraps(func)
        def _wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)
        setattr(cls, name, _wrapper)
        return func
    return _decorator


def add_progress_cancel_widget(func):  # pragma: no cover
    """
    Decorator that adds progress widget and cancel button to func that returns a RequestTask.

    :param funcion func: the async function to decorate
    :return: the decorator
    :rtype: decorator
    """
    @wraps(func)
    def _wrapper(self, *args, **kwargs):  # pylint: disable=too-many-locals
        result = func(self, *args, **kwargs)

        if isinstance(result, rockets.RequestTask) and in_notebook():
            from ipywidgets import FloatProgress, Label, HBox, VBox, Button
            from IPython.display import display

            progress = FloatProgress(min=0, max=1, value=0)
            label = Label(value='')
            button = Button(description='Cancel')
            box = VBox([label, HBox([progress, button])])
            display(box)

            def _on_cancel(value):  # pylint: disable=unused-argument
                result.cancel()

            def _on_progress(value):
                progress.value = value.amount
                label.value = value.operation

            def _on_done(task):  # pylint: disable=unused-argument
                box.close()

            button.on_click(_on_cancel)
            result.add_progress_callback(_on_progress)
            result.add_done_callback(_on_done)

        return result
    return _wrapper


def obtain_registry(url):
    """Obtain the registry of exposed objects and RPCs from Brayns."""
    status = http_request(HTTP_METHOD_GET, url, 'registry')
    if status.code != HTTP_STATUS_OK:
        raise Exception('Failed to obtain registry from Brayns')
    return status.contents


def build_schema_requests_from_registry(url):
    """Obtain the registry and return it alongside with a list of schema requests."""
    registry = obtain_registry(url)
    endpoints = {x.replace(SCHEMA_ENDPOINT, '') for x in registry}

    request_list = list()
    for endpoint in endpoints:
        request_list.append(rockets.Request('schema', {'endpoint': endpoint}))
    return registry, request_list


def convert_snapshot_response_to_PIL(response):
    """Convert the snapshot response from Brayns to a PIL image"""
    if not response:  # pragma: no cover
        return None

    # error case: invalid request/parameters
    if 'code' in response:
        print(response['message'])
        return None
    return Image.open(io.BytesIO(base64decode(response['data'])))


def base64decode(data):
    """Properly decode the given base64 string"""
    # https://stackoverflow.com/a/9807138
    missing_padding = len(data) % 4
    if missing_padding != 0:
        data += b'=' * (4 - missing_padding)
    return base64.b64decode(data)
